import tempfile
import uuid
from datetime import date, datetime

import pandas as pd
import pandas.testing as pd_testing
import pytz

from ..kawa_base_e2e_test import KawaBaseTest
from ...client.kawa_client import KawaClient


def utc(year, month, day, hour, minute, second):
    return datetime(year, month, day, hour, minute, second, tzinfo=pytz.UTC)


class TestComputationDsl(KawaBaseTest):

    @classmethod
    def setUpClass(cls):
        unique_id = 'resource_{}'.format(uuid.uuid4())
        print(tempfile.gettempdir())
        KawaBaseTest.setUpClass()
        cls.data, cls.df = cls.df_for_tests()
        cls.sheet_name = unique_id
        loader = cls.kawa.new_data_loader(df=cls.df.copy(deep=True), datasource_name=unique_id)
        loader.create_datasource()
        loader.load_data(reset_before_insert=True, create_sheet=True)

    def test_select_all_fields(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['text']), self.data['text'][0:5])
        self.assertListEqual(list(computed_df['boolean']), self.data['boolean'][0:5])
        self.assertListEqual(list(computed_df['integer']), self.data['integer'][0:5])
        self.assertListEqual(list(computed_df['decimal']), self.data['decimal'][0:5])
        self.assertListEqual(list(computed_df['date']), self.data['date'][0:5])
        self.assertListEqual(list(computed_df['date_time']), self.data['date_time'][0:5])

    def test_select_all_fields_preserves_the_order(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .select('text', 'date', 'boolean', 'integer', 'date_time', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df.columns), ['text', 'date', 'boolean', 'integer', 'date_time', 'decimal'])

    def test_select_multiple_fields_with_regexp(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .select(KawaClient.cols(r'.*date.*'))
                       .compute())

        # then
        self.assertListEqual(list(computed_df['date']), self.data['date'][0:5])
        self.assertListEqual(list(computed_df['date_time']), self.data['date_time'][0:5])

    def test_select_indicator_fields_only(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name, force_tz='UTC')
                       .select(KawaClient.indicator_cols())
                       .compute())

        pd_testing.assert_frame_equal(computed_df.drop('record_id', axis=1).sort_index(axis=1),
                                      self.df.sort_index(axis=1),
                                      check_dtype=False)
        # then
        self.assertListEqual(list(computed_df['date']), self.data['date'][0:5])
        self.assertListEqual(list(computed_df['date_time']), self.data['date_time'][0:5])

    def test_select_all_fields_with_gt_integer_filter(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('integer').gt(2))
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['text']), self.data['text'][2:5])
        self.assertListEqual(list(computed_df['integer']), self.data['integer'][2:5])
        self.assertListEqual(list(computed_df['decimal']), self.data['decimal'][2:5])
        self.assertListEqual(list(computed_df['date']), self.data['date'][2:5])
        self.assertListEqual(list(computed_df['date_time']), self.data['date_time'][2:5])

    def test_select_all_fields_with_lt_integer_filter(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('integer').lt(2))
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['text']), self.data['text'][0:1])
        self.assertListEqual(list(computed_df['integer']), self.data['integer'][0:1])
        self.assertListEqual(list(computed_df['decimal']), self.data['decimal'][0:1])
        self.assertListEqual(list(computed_df['date']), self.data['date'][0:1])
        self.assertListEqual(list(computed_df['date_time']), self.data['date_time'][0:1])

    def test_select_all_fields_with_gte_integer_filter(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('integer').gte(2))
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['text']), self.data['text'][1:5])
        self.assertListEqual(list(computed_df['integer']), self.data['integer'][1:5])
        self.assertListEqual(list(computed_df['decimal']), self.data['decimal'][1:5])
        self.assertListEqual(list(computed_df['date']), self.data['date'][1:5])
        self.assertListEqual(list(computed_df['date_time']), self.data['date_time'][1:5])

    def test_select_all_fields_with_lte_integer_filter(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('integer').lte(2))
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['text']), self.data['text'][0:2])
        self.assertListEqual(list(computed_df['integer']), self.data['integer'][0:2])
        self.assertListEqual(list(computed_df['decimal']), self.data['decimal'][0:2])
        self.assertListEqual(list(computed_df['date']), self.data['date'][0:2])
        self.assertListEqual(list(computed_df['date_time']), self.data['date_time'][0:2])

    def test_select_all_fields_with_eq_integer_filter(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('integer').eq(2))
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['text']), self.data['text'][1:2])
        self.assertListEqual(list(computed_df['integer']), self.data['integer'][1:2])
        self.assertListEqual(list(computed_df['decimal']), self.data['decimal'][1:2])
        self.assertListEqual(list(computed_df['date']), self.data['date'][1:2])
        self.assertListEqual(list(computed_df['date_time']), self.data['date_time'][1:2])

    def test_select_all_fields_with_ne_integer_filter(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('integer').ne(5))
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['text']), self.data['text'][0:4])
        self.assertListEqual(list(computed_df['integer']), self.data['integer'][0:4])
        self.assertListEqual(list(computed_df['decimal']), self.data['decimal'][0:4])
        self.assertListEqual(list(computed_df['date']), self.data['date'][0:4])
        self.assertListEqual(list(computed_df['date_time']), self.data['date_time'][0:4])

    def test_select_all_fields_with_starts_with_filter(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('text').starts_with('value1'))
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['text']), self.data['text'][0:2])
        self.assertListEqual(list(computed_df['integer']), self.data['integer'][0:2])
        self.assertListEqual(list(computed_df['decimal']), self.data['decimal'][0:2])
        self.assertListEqual(list(computed_df['date']), self.data['date'][0:2])
        self.assertListEqual(list(computed_df['date_time']), self.data['date_time'][0:2])

    def test_select_all_fields_with_ends_with_filter(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('text').ends_with('3'))
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['text']), self.data['text'][4:5])
        self.assertListEqual(list(computed_df['integer']), self.data['integer'][4:5])
        self.assertListEqual(list(computed_df['decimal']), self.data['decimal'][4:5])
        self.assertListEqual(list(computed_df['date']), self.data['date'][4:5])
        self.assertListEqual(list(computed_df['date_time']), self.data['date_time'][4:5])

    def test_select_all_fields_with_contains_filter(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('text').contains('e1'))
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['text']), self.data['text'][0:2])
        self.assertListEqual(list(computed_df['integer']), self.data['integer'][0:2])
        self.assertListEqual(list(computed_df['decimal']), self.data['decimal'][0:2])
        self.assertListEqual(list(computed_df['date']), self.data['date'][0:2])
        self.assertListEqual(list(computed_df['date_time']), self.data['date_time'][0:2])

    def test_select_all_fields_with_does_no_contain_filter(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('text').does_not_contain('e1'))
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['text']), self.data['text'][2:5])
        self.assertListEqual(list(computed_df['integer']), self.data['integer'][2:5])
        self.assertListEqual(list(computed_df['decimal']), self.data['decimal'][2:5])
        self.assertListEqual(list(computed_df['date']), self.data['date'][2:5])
        self.assertListEqual(list(computed_df['date_time']), self.data['date_time'][2:5])

    def test_select_all_fields_with_range_filter_on_date(self):
        # when
        d = date(2023, 1, 1)
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('date').date_range(from_inclusive=d, to_inclusive=d))
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['integer']), [self.data['integer'][0], self.data['integer'][2]])
        self.assertListEqual(list(computed_df['date']), [self.data['date'][0], self.data['date'][2]])

    def test_select_all_fields_with_range_filter_on_datetime(self):
        # when
        from_dt = utc(2023, 1, 1, 23, 23, 2)
        to_dt = utc(2023, 1, 1, 23, 23, 2)
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('date_time').datetime_range(from_inclusive=from_dt, to_inclusive=to_dt))
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['integer']), [
            self.data['integer'][0],
            self.data['integer'][2],
            self.data['integer'][3],
            self.data['integer'][4]])

    def test_select_all_fields_with_in_list_filter(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('text').in_list(['value11', 'value21']))
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['integer']), [self.data['integer'][0], self.data['integer'][2]])

    def test_select_all_fields_with_open_range_filter_on_datetime(self):
        # when
        from_dt = utc(2023, 1, 1, 23, 23, 3)
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('date_time').datetime_range(from_inclusive=from_dt))
                       .select('date', 'date_time', 'boolean', 'text', 'integer', 'decimal')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['integer']), [self.data['integer'][1]])

    def test_aggregation_without_filters(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .group_by('date')
                       .select(KawaClient.col('integer').avg(),
                               KawaClient.col('date').first(),
                               KawaClient.col('decimal').sum())
                       .compute()).sort_values(by=['date'])

        # then
        self.assertListEqual(list(computed_df.columns), ['grouping(0)', 'integer', 'date', 'decimal'])

        self.assertListEqual(list(computed_df['grouping(0)']), [
            date(2023, 1, 1),
            date(2023, 1, 2)
        ])
        self.assertListEqual(list(computed_df['date']), [
            date(2023, 1, 1),
            date(2023, 1, 2)
        ])

        self.assertListEqual(list(computed_df['integer']), [2.0, 11.0 / 3])
        self.assertListEqual(list(computed_df['decimal']), [1.1 + 3.1, 2.1 + 4.1 + 5.1])

    def test_aggregation_without_filters_and_without_limit(self):
        # when
        computed_df = ((self.kawa
                        .sheet(sheet_name=self.sheet_name)
                        .group_by('date')
                        .agg(KawaClient.col('integer').avg(),
                             KawaClient.col('date').first(),
                             KawaClient.col('decimal').sum())
                        .no_limit()
                        .compute())
                       .sort_values(by=['date']))

        # then
        self.assertListEqual(list(computed_df.columns), ['grouping(0)', 'integer', 'date', 'decimal'])
        self.assertListEqual(list(computed_df['grouping(0)']), [date(2023, 1, 1), date(2023, 1, 2)])
        self.assertListEqual(list(computed_df['date']), [date(2023, 1, 1), date(2023, 1, 2)])
        self.assertListEqual(list(computed_df['integer']), [2.0, 11.0 / 3])
        self.assertListEqual(list(computed_df['decimal']), [1.1 + 3.1, 2.1 + 4.1 + 5.1])

    def test_with_in_list_filter(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .filter(KawaClient.col('text').in_list(['value11', 'value21']))
                       .select('text')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['text']), ['value11', 'value21'])

    def test_with_in_list_filter_(self):
        # expect
        self.assertRaisesRegex(
            TypeError,
            'Type mismatch for typed variable',
            lambda: KawaClient.col('text').in_list('value11', 'value21'),
        )

    @staticmethod
    def df_for_tests():
        data = {
            'date': [
                date(2023, 1, 1),
                date(2023, 1, 2),
                date(2023, 1, 1),
                date(2023, 1, 2),
                date(2023, 1, 2)
            ],
            'date_time': [
                utc(2023, 1, 1, 23, 23, 2),
                utc(2023, 1, 1, 23, 23, 4),
                utc(2023, 1, 1, 23, 23, 2),
                utc(2023, 1, 1, 23, 23, 2),
                utc(2023, 1, 1, 23, 23, 2),
            ],
            'text': ['value11', 'value12', 'value21', 'value22', 'value23'],
            'integer': [1, 2, 3, 4, 5],
            'decimal': [1.1, 2.1, 3.1, 4.1, 5.1],
            'boolean': [True, False, True, True, False],
        }

        data_for_df = []
        for i in range(5):
            data_for_df.append({
                'date': data['date'][i],
                'date_time': data['date_time'][i],
                'boolean': data['boolean'][i],
                'text': data['text'][i],
                'integer': data['integer'][i],
                'decimal': data['decimal'][i],
            })

        return data, pd.DataFrame(data_for_df)



from ..kawa_base_e2e_test import KawaBaseTest


class TestReporting(KawaBaseTest):

    def test_workspace_member_reporting(self):
        r = self.kawa.reporting().generate_workspace_member_list_report()
        self.assertListEqual(list(r.columns), [
            'workspace',
            'name',
            'email',
            'unique_id',
            'is_admin',
            'can_see_all_data',
            'can_manage_security',
            'can_manage_users',
            'workspace_permissions'
        ])

    def test_principal_reporting(self):
        r = self.kawa.reporting().generate_user_list_report()
        print(list(r.columns))
        self.assertListEqual(list(r.columns), [
            'internal_id',
            'email',
            'name',
            'unique_id',
            'forbidden_data_source_types',
            'is_admin',
            'status',
            'global_permissions'
        ])



import pandas as pd

from kywy.client.kawa_client import KawaClient
from kywy.client.kawa_decorators import inputs, outputs, secrets


@inputs(measure1=float)
@outputs(measure3=float)
@secrets(api='api')
def execute(df: pd.DataFrame, kawa: KawaClient, api, missing_param):
    print('run: simple_join_script')
    print('[1]: Data from KAWA')
    print(df)
    kawa.get('')
    df['measure3'] = df['measure1'] * 5
    print('[2]: Data after script operation')
    print(df)
    return df



import pandas as pd

from kywy.client.kawa_decorators import outputs, inputs


@inputs(measure1=float, measure2=float)
@outputs(measure3=float)
def toto(df: pd.DataFrame):
    print(df)
    df['measure3'] = df.apply(lambda row: func(row['measure1'], row['measure2']), axis=1)
    return df


def func(a: float, b: float):
    return a + b



import pandas as pd


class KawaReporting:

    def __init__(self, kawa_client):
        self._k = kawa_client

    def generate_user_list_report(self):
        kawa = self._k
        principals = kawa.entities.principals().list_entities()
        principal_descriptions = []
        for principal in principals:
            principal_descriptions.append({
                'internal_id': principal.get('id'),
                'email': principal.get('email'),
                'name': principal.get('displayInformation').get('displayName'),
                'unique_id': principal.get('uniqueId'),
                'forbidden_data_source_types': principal.get('forbiddenDataSourceTypes'),
                'is_admin': principal.get('role') != 'READER_ROLE',
                'status': principal.get('status'),
                'global_permissions': principal.get('permissions'),
            })

        return pd.DataFrame(principal_descriptions)

    def generate_workspace_member_list_report(self):
        kawa = self._k
        workspaces = kawa.entities.workspaces().list_entities()
        principals = kawa.entities.principals().list_entities()

        principal_descriptions = {}
        for principal in principals:
            principal_description = {
                'internal_id': principal.get('id'),
                'email': principal.get('email'),
                'name': principal.get('displayInformation').get('displayName'),
                'unique_id': principal.get('uniqueId'),
                'forbidden_data_source_types': principal.get('forbiddenDataSourceTypes'),
                'is_admin': principal.get('role') != 'READER_ROLE',
                'status': principal.get('status'),
                'global_permissions': principal.get('permissions'),
            }
            principal_descriptions[principal.get('id')] = principal_description

        members = []
        for workspace in workspaces:
            for member in workspace.get('members'):
                member_id = member.get('principalId')
                member_description = principal_descriptions.get(member_id)
                is_admin = member_description.get('is_admin', False)
                members.append({
                    'workspace': workspace.get('displayInformation').get('displayName'),
                    'name': member_description.get('name'),
                    'email': member_description.get('email'),
                    'unique_id': member_description.get('unique_id'),
                    'is_admin': member_description.get('is_admin'),
                    'can_see_all_data': is_admin or 'MANAGE_DATA_SOURCES' in member.get('permissions'),
                    'can_manage_security': is_admin or 'ROW_LEVEL_SECURITY_POLICY' in member.get('permissions'),
                    'can_manage_users': is_admin or 'MANAGE_USERS' in member.get('permissions'),
                    'workspace_permissions': ','.join(member.get('permissions')),
                })

        return pd.DataFrame(members)



from ...kawa_base_e2e_test import KawaBaseTest
import pandas as pd
import uuid
import tempfile
from ....client.kawa_client import KawaClient as K


class TestComputationWithTextFilters(KawaBaseTest):

    @classmethod
    def setUpClass(cls):
        unique_id = 'resource_{}'.format(uuid.uuid4())
        print(tempfile.gettempdir())
        KawaBaseTest.setUpClass()
        cls.df = cls._df_for_tests()
        cls.sheet_name = unique_id
        loader = cls.kawa.new_data_loader(df=cls.df.copy(deep=True), datasource_name=unique_id)
        loader.create_datasource()
        loader.load_data(reset_before_insert=True, create_sheet=True)

    def test_without_filters(self):
        # when
        computed_df = (self.kawa
                       .sheet(sheet_name=self.sheet_name)
                       .select('text')
                       .compute())

        # then
        self.assertListEqual(list(computed_df['text']), ['this', 'is', 'some', 'test', 'data', ''])

    def test_contains_filter(self):
        self._check_filter(K.col('text').contains('i'), ['this', 'is'])
        self._check_filter(K.col('text').contains('data'), ['data'])
        self._check_filter(K.col('text').contains('foo'), [])
        self._check_filter(K.col('text').contains(''), ['this', 'is', 'some', 'test', 'data', ''])

    def test_does_not_contain_filter(self):
        self._check_filter(K.col('text').does_not_contain('i'), ['some', 'test', 'data', ''])
        self._check_filter(K.col('text').does_not_contain('data'), ['this', 'is', 'some', 'test', ''])
        self._check_filter(K.col('text').does_not_contain('foo'), ['this', 'is', 'some', 'test', 'data', ''])
        self._check_filter(K.col('text').does_not_contain(''), [])

    def test_starts_with(self):
        self._check_filter(K.col('text').starts_with('is'), ['is'])
        self._check_filter(K.col('text').starts_with('data'), ['data'])
        self._check_filter(K.col('text').starts_with('foo'), [])
        self._check_filter(K.col('text').starts_with(''), ['this', 'is', 'some', 'test', 'data', ''])

    def test_does_not_start_with(self):
        self._check_filter(K.col('text').does_not_start_with('is'), ['this', 'some', 'test', 'data', ''])
        self._check_filter(K.col('text').does_not_start_with('data'), ['this', 'is', 'some', 'test', ''])
        self._check_filter(K.col('text').does_not_start_with('foo'),
                           ['this', 'is', 'some', 'test', 'data', ''])
        self._check_filter(K.col('text').does_not_start_with(''), [])

    def test_ends_with(self):
        self._check_filter(K.col('text').ends_with('s'), ['this', 'is'])
        self._check_filter(K.col('text').ends_with('is'), ['this', 'is'])
        self._check_filter(K.col('text').ends_with('his'), ['this'])
        self._check_filter(K.col('text').ends_with('data'), ['data'])
        self._check_filter(K.col('text').ends_with('foo'), [])
        self._check_filter(K.col('text').ends_with(''), ['this', 'is', 'some', 'test', 'data', ''])

    def test_does_not_end_with(self):
        self._check_filter(K.col('text').does_not_end_with('s'), ['some', 'test', 'data', ''])
        self._check_filter(K.col('text').does_not_end_with('is'), ['some', 'test', 'data', ''])
        self._check_filter(K.col('text').does_not_end_with('his'), ['is', 'some', 'test', 'data', ''])
        self._check_filter(K.col('text').does_not_end_with('data'), ['this', 'is', 'some', 'test', ''])
        self._check_filter(K.col('text').does_not_end_with('foo'), ['this', 'is', 'some', 'test', 'data', ''])
        self._check_filter(K.col('text').does_not_end_with(''), [])

    def test_empty(self):
        self._check_filter(K.col('text').empty(), [''])

    def test_not_empty(self):
        self._check_filter(K.col('text').not_empty(), ['this', 'is', 'some', 'test', 'data'])

    def test_in_list(self):
        self._check_filter(K.col('text').in_list(['this', 'is']), ['this', 'is'])
        self._check_filter(K.col('text').in_list(['this', 'is', 'foo']), ['this', 'is'])
        self._check_filter(K.col('text').in_list(['this', 'is', '']), ['this', 'is', ''])
        self._check_filter(K.col('text').in_list(['foo']), [])
        self._check_filter(K.col('text').in_list(['']), [''])

    def test_not_in_list(self):
        self._check_filter(K.col('text').in_list(['this', 'is']).exclude(), ['some', 'test', 'data', ''])
        self._check_filter(K.col('text').in_list(['foo']).exclude(), ['this', 'is', 'some', 'test', 'data', ''])

    def _check_filter(self, f, expected):
        df = (self.kawa
              .sheet(sheet_name=self.sheet_name)
              .filter(f)
              .select('text')
              .compute())
        self.assertListEqual(list(df['text']), expected)

    @staticmethod
    def _df_for_tests():
        data = {
            'text': ['this', 'is', 'some', 'test', 'data', '']
        }
        return pd.DataFrame(data)



import logging

import pandas as pd

from kywy.client.kawa_decorators import inputs, outputs


@inputs(measure1=float, measure2=float)
@outputs(measure3=float)
def toto(df: pd.DataFrame):
    script_logger = logging.getLogger('script-logger')
    script_logger.info('Computing the df hahahahahahahahah')
    df['measure3'] = df['measure1'] + df['measure2']
    script_logger.info('Returning the df')
    script_logger.info(df)
    return df



import pandas as pd
from datetime import date, datetime
import zoneinfo
import uuid
import math
import time

from ..kawa_base_e2e_test import KawaBaseTest


class TestDataLoader(KawaBaseTest):

    def test_that_datasource_is_properly_defined(self):
        # setup
        loader = self.kawa.new_data_loader(
            df=pd.DataFrame([{'boolean1': True, 'text1': 'bar', 'measure1': 1.124}]),
            datasource_name=' Orders')

        # when
        ds = loader._define_data_source_from_df(primary_keys=['text1'])

        # then
        self.assertFalse(ds['shared'])
        self.assertEqual(ds['displayInformation']['displayName'], 'Orders')
        self.assertEqual(ds['storageConfiguration']['loadingAdapterName'], 'CLICKHOUSE')
        self.assertEqual(len(ds['indicators']), 3)

        self.assertEqual(ds['indicators'][0]['displayInformation']['displayName'], 'boolean1')
        self.assertEqual(ds['indicators'][0]['indicatorId'], 'boolean1')
        self.assertEqual(ds['indicators'][0]['type'], 'boolean')
        self.assertIsNone(ds['indicators'][0].get('key'))
        self.assertFalse(ds['indicators'][0]['storageConfig']['indexed'])

        self.assertEqual(ds['indicators'][1]['displayInformation']['displayName'], 'text1')
        self.assertEqual(ds['indicators'][1]['indicatorId'], 'text1')
        self.assertEqual(ds['indicators'][1]['type'], 'text')
        self.assertEqual(ds['indicators'][1]['key']['keyType'], 'PRIMARY_SHARDING_KEY')
        self.assertTrue(ds['indicators'][1]['storageConfig']['indexed'])

        self.assertEqual(ds['indicators'][2]['displayInformation']['displayName'], 'measure1')
        self.assertEqual(ds['indicators'][2]['indicatorId'], 'measure1')
        self.assertEqual(ds['indicators'][2]['type'], 'decimal')
        self.assertIsNone(ds['indicators'][2].get('key'))
        self.assertFalse(ds['indicators'][2]['storageConfig']['indexed'])

    def test_that_an_auto_increment_is_added(self):
        # setup
        loader = self.kawa.new_data_loader(
            df=pd.DataFrame([{'boolean1': True, 'text1': 'bar', 'measure1': 1.124}]),
            datasource_name='Orders')

        # when
        ds = loader._define_data_source_from_df()

        # then
        self.assertFalse(ds['shared'])
        self.assertEqual(ds['displayInformation']['displayName'], 'Orders')
        self.assertEqual(ds['storageConfiguration']['loadingAdapterName'], 'CLICKHOUSE')
        self.assertEqual(len(ds['indicators']), 4)

        self.assertEqual(ds['indicators'][0]['displayInformation']['displayName'], 'record_id')
        self.assertEqual(ds['indicators'][0]['indicatorId'], 'record_id')
        self.assertEqual(ds['indicators'][0]['type'], 'integer')
        self.assertIsNotNone(ds['indicators'][0].get('key'))
        self.assertEqual(ds['indicators'][0]['key']['keyType'], 'PRIMARY_SHARDING_KEY')
        self.assertTrue(ds['indicators'][0]['storageConfig']['indexed'])

        self.assertEqual(ds['indicators'][1]['displayInformation']['displayName'], 'boolean1')
        self.assertEqual(ds['indicators'][1]['indicatorId'], 'boolean1')
        self.assertEqual(ds['indicators'][1]['type'], 'boolean')
        self.assertIsNone(ds['indicators'][1].get('key'))
        self.assertFalse(ds['indicators'][1]['storageConfig']['indexed'])

        self.assertEqual(ds['indicators'][2]['displayInformation']['displayName'], 'text1')
        self.assertEqual(ds['indicators'][2]['indicatorId'], 'text1')
        self.assertEqual(ds['indicators'][2]['type'], 'text')
        self.assertIsNone(ds['indicators'][2].get('key'))
        self.assertFalse(ds['indicators'][2]['storageConfig']['indexed'])

        self.assertEqual(ds['indicators'][3]['displayInformation']['displayName'], 'measure1')
        self.assertEqual(ds['indicators'][3]['indicatorId'], 'measure1')
        self.assertEqual(ds['indicators'][3]['type'], 'decimal')
        self.assertIsNone(ds['indicators'][3].get('key'))
        self.assertFalse(ds['indicators'][3]['storageConfig']['indexed'])

    def test_that_array_types_are_correctly_introspected(self):
        # setup
        df = pd.DataFrame([
            {
                'array_of_texts_1': None,
                'array_of_texts_2': ['foo'],
                'array_of_decimals_1': None,
            },
            {
                'array_of_texts_1': [],
                'array_of_texts_2': [1],
                'array_of_decimals_1': [1.0, 2.0, 3.0],
            },
            {
                'array_of_texts_1': [None, 'bar'],
                'array_of_texts_3': ['foo', 1],
                'array_of_integers_1': [1],
            },
        ])
        loader = self.kawa.new_data_loader(df=df, datasource_name='Foo')

        # when
        introspection = loader._introspect_df()

        # then
        self.assertEqual(introspection['array_of_texts_1'], 'list(integer,text)')
        self.assertEqual(introspection['array_of_texts_2'], 'list(integer,text)')
        self.assertEqual(introspection['array_of_texts_3'], 'list(integer,text)')
        self.assertEqual(introspection['array_of_decimals_1'], 'list(integer,decimal)')
        self.assertEqual(introspection['array_of_integers_1'], 'list(integer,integer)')

    def test_that_basic_types_are_correctly_introspected(self):
        # setup
        df = pd.DataFrame([
            {'boolean': True},
            {'boolean': True, 'text1': 'bar'},
            {'boolean': True},
            {'boolean': True, 'text1': None},
        ])
        df['text2'] = df['text1'].astype('string')
        loader = self.kawa.new_data_loader(df=df, datasource_name='Foo')

        # when
        introspection = loader._introspect_df()

        # then
        self.assertEqual(introspection['boolean'], 'boolean')
        self.assertEqual(introspection['text1'], 'text')
        self.assertEqual(introspection['text2'], 'text')

    def test_that_numeric_types_are_correctly_introspected(self):
        # setup
        df = pd.DataFrame([{'int': 12, 'float': 1.23}])
        df['int8'] = df['int'].astype('int8')
        df['int16'] = df['int'].astype('int16')
        df['int32'] = df['int'].astype('int32')
        df['int64'] = df['int'].astype('int64')
        df['uint8'] = df['int'].astype('uint8')
        df['uint16'] = df['int'].astype('uint16')
        df['uint32'] = df['int'].astype('uint32')
        df['uint64'] = df['int'].astype('uint64')
        df['float16'] = df['float'].astype('float16')
        df['float32'] = df['float'].astype('float32')
        df['float64'] = df['float'].astype('float64')
        loader = self.kawa.new_data_loader(df=df, datasource_name='Foo')

        # when
        introspection = loader._introspect_df()

        # then
        self.assertEqual(introspection['int'], 'integer')
        self.assertEqual(introspection['int8'], 'integer')
        self.assertEqual(introspection['int16'], 'integer')
        self.assertEqual(introspection['int32'], 'integer')
        self.assertEqual(introspection['int64'], 'integer')
        self.assertEqual(introspection['uint8'], 'integer')
        self.assertEqual(introspection['uint16'], 'integer')
        self.assertEqual(introspection['uint32'], 'integer')
        self.assertEqual(introspection['uint64'], 'integer')
        self.assertEqual(introspection['float'], 'decimal')
        self.assertEqual(introspection['float16'], 'decimal')
        self.assertEqual(introspection['float32'], 'decimal')
        self.assertEqual(introspection['float64'], 'decimal')

    def test_that_temporal_types_are_correctly_introspected(self):
        # setup
        pacific = zoneinfo.ZoneInfo('US/Pacific')
        utc = zoneinfo.ZoneInfo('UTC')
        df = pd.DataFrame([{
            # DATE TYPE
            'date': date(2023, 1, 1),
            # DATE TIME TYPE
            'date_time_us_pacific': datetime(2023, 1, 1, 23, 23, 2, tzinfo=pacific),
            'date_time_utc': datetime(2023, 1, 1, 23, 23, 2, tzinfo=utc),
            'date_time_no_tz': datetime(2023, 1, 1, 23, 23, 2),
            # TIMESTAMP TYPE
            'ts_us_pacific': pd.Timestamp(year=2023, month=1, day=1, hour=23, minute=23, second=2, tz='US/Pacific'),
            'ts_utc': pd.Timestamp(year=2023, month=1, day=1, hour=23, minute=23, second=2, tz='utc'),
            'ts_no_tz': pd.Timestamp(year=2023, month=1, day=1, hour=23, minute=23, second=2),
        }])
        loader = self.kawa.new_data_loader(df=df, datasource_name='Foo')

        # when
        introspection = loader._introspect_df()

        # then
        self.assertEqual(introspection['date'], 'date')
        self.assertEqual(introspection['date_time_us_pacific'], 'date_time')
        self.assertEqual(introspection['date_time_utc'], 'date_time')
        self.assertEqual(introspection['date_time_no_tz'], 'date_time')
        self.assertEqual(introspection['ts_us_pacific'], 'date_time')
        self.assertEqual(introspection['ts_utc'], 'date_time')
        self.assertEqual(introspection['ts_no_tz'], 'date_time')

    def test_that_ingestion_without_primary_key_correctly_uses_auto_increment(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {'boolean1': True, 'text1': 'bar', 'integer1': 1},
            {'boolean1': True, 'text1': 'bar', 'integer1': 1},
            {'boolean1': True, 'text1': 'bar', 'integer1': 1},
        ])
        loader = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader.create_datasource()

        # when
        loader.load_data(reset_before_insert=True, create_sheet=True)
        loader.load_data(reset_before_insert=False)
        loader.load_data(reset_before_insert=False)

        # then
        computed_df = (self.kawa
                       .sheet(sheet_name=unique_id)
                       .select('boolean1', 'text1', 'integer1', 'record_id')
                       .compute())
        self.assertEqual(computed_df.shape[0], 9)
        for row_id in range(9):
            self.assertTrue(computed_df['boolean1'][row_id])
            self.assertEqual(computed_df['text1'][row_id], 'bar')
            self.assertEqual(computed_df['integer1'][row_id], 1)
            self.assertEqual(computed_df['record_id'][row_id], row_id + 1)

    def test_that_optimize_after_insert_remove_duplicate_pk(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader1 = self.kawa.new_data_loader(
            df=pd.DataFrame([
                {'key': 'k1', 'value': 11},
                {'key': 'k2', 'value': 12},
                {'key': 'k3', 'value': 13},
            ]),
            datasource_name=unique_id
        )
        loader1.create_datasource(primary_keys=['key'])
        loader1.load_data(reset_before_insert=True, create_sheet=True)

        loader2 = self.kawa.new_data_loader(
            df=pd.DataFrame([
                {'key': 'k4', 'value': 24},
                {'key': 'k5', 'value': 25},
                {'key': 'k3', 'value': 23},
            ]),
            datasource_name=unique_id
        )

        # when
        loader2.load_data(reset_before_insert=False, optimize_after_insert=True)

        # then
        computed_df = (self.kawa
                       .sheet(sheet_name=unique_id)
                       .select('key', 'value')
                       .compute())
        self.assertEqual(computed_df.shape[0], 5)
        self.assertListEqual(list(computed_df['key']), ['k1', 'k2', 'k3', 'k4', 'k5'])
        self.assertListEqual(list(computed_df['value']), [11, 12, 23, 24, 25])

    def test_that_optimize_after_insert_does_not_remove_duplicate_pk_when_not_optimized(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader1 = self.kawa.new_data_loader(
            df=pd.DataFrame([
                {'key': 'k1', 'value': 11},
                {'key': 'k2', 'value': 12},
                {'key': 'k3', 'value': 13},
            ]),
            datasource_name=unique_id
        )
        loader1.create_datasource(primary_keys=['key'])
        loader1.load_data(reset_before_insert=True, create_sheet=True)

        loader2 = self.kawa.new_data_loader(
            df=pd.DataFrame([
                {'key': 'k4', 'value': 24},
                {'key': 'k5', 'value': 25},
                {'key': 'k3', 'value': 23},
            ]),
            datasource_name=unique_id
        )

        # when
        loader2.load_data(reset_before_insert=False)

        # then
        computed_df = (self.kawa
                       .sheet(sheet_name=unique_id)
                       .select('key', 'value')
                       .compute())
        self.assertEqual(computed_df.shape[0], 6)
        self.assertListEqual(list(computed_df['key']), ['k1', 'k2', 'k3', 'k3', 'k4', 'k5'])
        self.assertListEqual(list(computed_df['value']), [11, 12, 23, 13, 24, 25])

    def test_that_command_force_entries_deduplication_removes_duplicates(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader1 = self.kawa.new_data_loader(
            df=pd.DataFrame([
                {'key': 'k1', 'value': 11},
                {'key': 'k2', 'value': 12},
                {'key': 'k3', 'value': 13},
            ]),
            datasource_name=unique_id
        )
        ds = loader1.create_datasource(primary_keys=['key'])
        loader1.load_data(reset_before_insert=True, create_sheet=True)

        loader2 = self.kawa.new_data_loader(
            df=pd.DataFrame([
                {'key': 'k4', 'value': 24},
                {'key': 'k5', 'value': 25},
                {'key': 'k3', 'value': 23},
            ]),
            datasource_name=unique_id
        )

        # when
        loader2.load_data(reset_before_insert=False)

        # then
        computed_df = (self.kawa
                       .sheet(sheet_name=unique_id)
                       .select('key', 'value')
                       .compute())
        self.assertEqual(computed_df.shape[0], 6)
        self.assertListEqual(list(computed_df['key']), ['k1', 'k2', 'k3', 'k3', 'k4', 'k5'])
        self.assertListEqual(list(computed_df['value']), [11, 12, 23, 13, 24, 25])

        # after
        self.kawa.commands.force_entries_deduplication(datasource=ds)

        # then
        computed_df = (self.kawa
                       .sheet(sheet_name=unique_id)
                       .select('key', 'value')
                       .compute())
        self.assertEqual(computed_df.shape[0], 5)
        self.assertListEqual(list(computed_df['key']), ['k1', 'k2', 'k3', 'k4', 'k5'])
        self.assertListEqual(list(computed_df['value']), [11, 12, 23, 24, 25])

    def test_that_reset_data_clears_data_between_each_load(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {'boolean1': True, 'text1': 'bar', 'integer1': 1},
            {'boolean1': True, 'text1': 'bar', 'integer1': 1},
            {'boolean1': True, 'text1': 'bar', 'integer1': 1},
        ])
        loader = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader.create_datasource()

        # when
        loader.load_data(reset_before_insert=True, create_sheet=True)
        loader.load_data(reset_before_insert=True)
        loader.load_data(reset_before_insert=True)

        # then
        computed_df = (self.kawa
                       .sheet(sheet_name=unique_id)
                       .select('boolean1', 'text1', 'integer1', 'record_id')
                       .compute())
        self.assertEqual(computed_df.shape[0], 3)
        for row_id in range(3):
            self.assertTrue(computed_df['boolean1'][row_id])
            self.assertEqual(computed_df['text1'][row_id], 'bar')
            self.assertEqual(computed_df['integer1'][row_id], 1)
            self.assertEqual(computed_df['record_id'][row_id], row_id + 1)

    def test_that_ingestion_works_when_omitting_columns(self):
        # setup
        utc = zoneinfo.ZoneInfo('UTC')
        unique_id = 'resource_{}'.format(uuid.uuid4())
        row1 = pd.DataFrame([{
            'boolean': True,
            'text': 'bar',
            'integer': 1,
            'decimal': 3.14,
            'date': date(2023, 1, 1),
            'date_time': datetime(2023, 1, 1, 23, 23, 2),
        }])
        row2 = pd.DataFrame([{}])

        loader1 = self.kawa.new_data_loader(df=row1, datasource_name=unique_id)
        loader2 = self.kawa.new_data_loader(df=row2, datasource_name=unique_id)

        loader1.create_datasource()

        # when
        loader1.load_data(reset_before_insert=True, create_sheet=True)
        loader2.load_data(reset_before_insert=False)

        # then
        computed_df = (self.kawa
                       .sheet(sheet_name=unique_id)
                       .select('boolean', 'text', 'integer', 'decimal', 'date', 'date_time', 'record_id')
                       .compute())

        self.assertTrue(computed_df['boolean'][0])
        self.assertEqual(computed_df['text'][0], 'bar')
        self.assertEqual(computed_df['integer'][0], 1)
        self.assertEqual(computed_df['decimal'][0], 3.14)
        self.assertEqual(computed_df['date'][0], date(2023, 1, 1))
        self.assertEqual(computed_df['decimal'][0], 3.14)
        self.assertEqual(computed_df['date_time'][0].timestamp(),
                         datetime(2023, 1, 1, 23, 23, 2, tzinfo=utc).timestamp())

        self.assertIsNone(computed_df['boolean'][1])
        self.assertEqual(computed_df['text'][1], '')
        self.assertTrue(math.isnan(computed_df['integer'][1]))
        self.assertTrue(math.isnan(computed_df['decimal'][1]))
        self.assertEqual(computed_df['date'][1], date(1970, 1, 1))
        self.assertEqual(computed_df['date_time'][1].timestamp(), 0)

    def test_that_every_data_type_is_properly_ingested(self):
        # setup
        pacific = zoneinfo.ZoneInfo('US/Pacific')
        utc = zoneinfo.ZoneInfo('UTC')
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf = pd.DataFrame([{
            # Arrays
            # FIXME: Bug when computing arrays of texts, returns empty text values
            'array_of_texts': ['foo', 'bar'],
            'array_of_numbers': [1.1, -2.2],
            'array_of_integers': [1, -2],

            # Date
            'date': date(2023, 1, 1),  # No idea why this works perfectly - need to investigate

            # Date + Time
            "date_time_utc": datetime(2023, 1, 1, 23, 23, 2, tzinfo=utc),
            "date_time_us_pacific": datetime(2023, 1, 1, 23, 23, 2, tzinfo=pacific),
            "date_time_no_tz": datetime(2023, 1, 1, 23, 23, 2),

            "ts_us_pacific": pd.Timestamp(year=2023, month=1, day=1, hour=23, minute=23, second=2, tz='US/Pacific'),
            "ts_utc": pd.Timestamp(year=2023, month=1, day=1, hour=23, minute=23, second=2, tz='utc'),
            "ts_no_tz": pd.Timestamp(year=2023, month=1, day=1, hour=23, minute=23, second=2),

            # Basic types
            'boolean1': True,
            'text1': 'bar',
            'integer1': 1,
            'measure1': 4.4
        }])

        dtypes = dtf.dtypes.to_dict()

        m = ''
        for col_name, typ in dtypes.items():
            m += f'dataframe[{col_name}].dtype == {typ}\n'

        loader = self.kawa.new_data_loader(df=dtf, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['text1'])

        # when
        loader.load_data(reset_before_insert=False, create_sheet=True)

        # then
        computed_df = self.kawa.sheet(sheet_name=unique_id).select(
            'date',
            'boolean1',
            'text1',
            'measure1',
            'integer1',
            'array_of_texts',
            'array_of_numbers',
            'array_of_integers',
            'date_time_utc',
            'date_time_no_tz',
            'date_time_us_pacific',
            'ts_us_pacific',
            'ts_utc',
            'ts_no_tz').compute()

        self.assertTrue(computed_df['boolean1'][0])
        self.assertEqual(computed_df['text1'][0], 'bar')
        self.assertEqual(computed_df['measure1'][0], 4.4)
        self.assertEqual(computed_df['integer1'][0], 1)
        self.assertListEqual(list(computed_df['array_of_numbers'][0]), [1.1, -2.2])
        self.assertListEqual(list(computed_df['array_of_integers'][0]), [1, -2])
        self.assertEqual(computed_df['date'][0], date(2023, 1, 1))
        self.assertEqual(computed_df['date'][0], date(2023, 1, 1))
        self.assertEqual(computed_df['date_time_utc'][0].timestamp(),
                         datetime(2023, 1, 1, 23, 23, 2, tzinfo=utc).timestamp())
        self.assertEqual(computed_df['date_time_us_pacific'][0].timestamp(),
                         datetime(2023, 1, 1, 23, 23, 2, tzinfo=pacific).timestamp())
        self.assertEqual(computed_df['date_time_no_tz'][0].timestamp(),
                         datetime(2023, 1, 1, 23, 23, 2, tzinfo=utc).timestamp())
        self.assertEqual(computed_df['ts_utc'][0].timestamp(),
                         datetime(2023, 1, 1, 23, 23, 2, tzinfo=utc).timestamp())
        self.assertEqual(computed_df['ts_us_pacific'][0].timestamp(),
                         datetime(2023, 1, 1, 23, 23, 2, tzinfo=pacific).timestamp())
        self.assertEqual(computed_df['ts_no_tz'][0].timestamp(),
                         datetime(2023, 1, 1, 23, 23, 2, tzinfo=utc).timestamp())

    def test_that_data_is_properly_ingested_on_multiple_threads_with_auto_increment(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader1 = self.kawa.new_data_loader(
            df=pd.DataFrame([
                {'date': date(2024, 1, 1), 'int': 1},
                {'date': date(2024, 1, 1), 'int': 2},
                {'date': date(2024, 1, 1), 'int': 3},
                {'date': date(2024, 1, 2), 'int': 4},
                {'date': date(2024, 1, 2), 'int': 5},
                {'date': date(2024, 1, 3), 'int': 6},
                {'date': date(2024, 1, 3), 'int': 7},
            ]),
            datasource_name=unique_id)
        loader2 = self.kawa.new_data_loader(
            df=pd.DataFrame([
                {'date': date(2024, 1, 1), 'int': 8},
                {'date': date(2024, 1, 1), 'int': 9},
                {'date': date(2024, 1, 1), 'int': 10},
                {'date': date(2024, 1, 2), 'int': 11},
                {'date': date(2024, 1, 2), 'int': 12},
                {'date': date(2024, 1, 3), 'int': 13},
                {'date': date(2024, 1, 3), 'int': 14},
            ]),
            datasource_name=unique_id)
        loader1.create_datasource()

        # when
        loader1.load_data(nb_threads=3, create_sheet=True)
        loader2.load_data(nb_threads=4)

        # then
        computed_df = self.kawa.sheet(sheet_name=unique_id).select('date', 'int', 'record_id').compute()
        self.assertListEqual(list(computed_df['int']), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])
        self.assertListEqual(list(computed_df['record_id']), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])
        self.assertListEqual(list(computed_df['date']), [
            date(2024, 1, 1),
            date(2024, 1, 1),
            date(2024, 1, 1),
            date(2024, 1, 2),
            date(2024, 1, 2),
            date(2024, 1, 3),
            date(2024, 1, 3),
            date(2024, 1, 1),
            date(2024, 1, 1),
            date(2024, 1, 1),
            date(2024, 1, 2),
            date(2024, 1, 2),
            date(2024, 1, 3),
            date(2024, 1, 3)
        ])

    def test_that_data_is_properly_ingested_on_multiple_threads_without_auto_increment(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader1 = self.kawa.new_data_loader(
            df=pd.DataFrame([
                {'date': date(2024, 1, 1), 'int': 1},
                {'date': date(2024, 1, 1), 'int': 2},
                {'date': date(2024, 1, 1), 'int': 3},
                {'date': date(2024, 1, 2), 'int': 4},
                {'date': date(2024, 1, 2), 'int': 5},
                {'date': date(2024, 1, 3), 'int': 6},
                {'date': date(2024, 1, 3), 'int': 7},
            ]),
            datasource_name=unique_id)
        loader2 = self.kawa.new_data_loader(
            df=pd.DataFrame([
                {'date': date(2024, 1, 1), 'int': 8},
                {'date': date(2024, 1, 1), 'int': 9},
                {'date': date(2024, 1, 1), 'int': 10},
                {'date': date(2024, 1, 2), 'int': 11},
                {'date': date(2024, 1, 2), 'int': 12},
                {'date': date(2024, 1, 3), 'int': 13},
                {'date': date(2024, 1, 3), 'int': 14},
            ]),
            datasource_name=unique_id
        )
        loader1.create_datasource(primary_keys=['int'])

        # when
        loader1.load_data(nb_threads=3, create_sheet=True)
        loader2.load_data(nb_threads=4)

        # then
        computed_df = self.kawa.sheet(sheet_name=unique_id).select('date', 'int').compute()
        self.assertListEqual(list(computed_df['int']), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])
        self.assertListEqual(list(computed_df['date']), [
            date(2024, 1, 1),
            date(2024, 1, 1),
            date(2024, 1, 1),
            date(2024, 1, 2),
            date(2024, 1, 2),
            date(2024, 1, 3),
            date(2024, 1, 3),
            date(2024, 1, 1),
            date(2024, 1, 1),
            date(2024, 1, 1),
            date(2024, 1, 2),
            date(2024, 1, 2),
            date(2024, 1, 3),
            date(2024, 1, 3)
        ])

    def test_that_computation_results_change_after_new_loads(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {'boolean1': True, 'text1': 'bar', 'integer1': 1},
            {'boolean1': True, 'text1': 'bar', 'integer1': 1},
            {'boolean1': True, 'text1': 'bar', 'integer1': 1},
        ])
        loader = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader.create_datasource()

        # when: compute 1
        loader.load_data(reset_before_insert=True, create_sheet=True)
        computed_df_1 = (self.kawa
                         .sheet(sheet_name=unique_id)
                         .select('boolean1', 'text1', 'integer1', 'record_id')
                         .compute())

        # when: compute 2
        loader.load_data(reset_before_insert=False)
        time.sleep(1)
        computed_df_2 = (self.kawa
                         .sheet(sheet_name=unique_id)
                         .select('boolean1', 'text1', 'integer1', 'record_id')
                         .compute())

        # when: compute 3
        loader.load_data(reset_before_insert=False)
        time.sleep(1)
        computed_df_3 = (self.kawa
                         .sheet(sheet_name=unique_id)
                         .select('boolean1', 'text1', 'integer1', 'record_id')
                         .compute())

        # then
        self.assertEqual(computed_df_1.shape[0], 3)
        for row_id in range(3):
            self.assertTrue(computed_df_1['boolean1'][row_id])
            self.assertEqual(computed_df_1['text1'][row_id], 'bar')
            self.assertEqual(computed_df_1['integer1'][row_id], 1)
            self.assertEqual(computed_df_1['record_id'][row_id], row_id + 1)

        self.assertEqual(computed_df_2.shape[0], 6)
        for row_id in range(6):
            self.assertTrue(computed_df_2['boolean1'][row_id])
            self.assertEqual(computed_df_2['text1'][row_id], 'bar')
            self.assertEqual(computed_df_2['integer1'][row_id], 1)
            self.assertEqual(computed_df_2['record_id'][row_id], row_id + 1)

        self.assertEqual(computed_df_3.shape[0], 9)
        for row_id in range(9):
            self.assertTrue(computed_df_3['boolean1'][row_id])
            self.assertEqual(computed_df_3['text1'][row_id], 'bar')
            self.assertEqual(computed_df_3['integer1'][row_id], 1)
            self.assertEqual(computed_df_3['record_id'][row_id], row_id + 1)

    def test_that_reindexing_preserves_auto_increment_behaviour(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=pd.DataFrame([
            {'text': 'bar', 'int': 1},
            {'text': 'foo', 'int': 2},
            {'text': 'bar', 'int': 3},
        ]), datasource_name=unique_id)
        ds = loader.create_datasource()
        loader.load_data(reset_before_insert=True, create_sheet=True)
        schema_1 = self.kawa.entities.get_datasource_schema(datasource_id=ds.get('id'))

        # when: change key and compute
        self.kawa.commands.replace_datasource_primary_keys(datasource=ds,
                                                           new_primary_keys=['text', 'record_id'])
        schema_2 = self.kawa.entities.get_datasource_schema(datasource_id=ds.get('id'))

        computed_df1 = (self.kawa
                        .sheet(sheet_name=unique_id)
                        .select('text', 'record_id', 'int')
                        .compute())

        # and: feed and compute once more
        loader.load_data(reset_before_insert=False, create_sheet=False)
        loader.load_data(reset_before_insert=False, create_sheet=False)
        time.sleep(5)  # Need time for cache to clear ... this is eventually consistent
        computed_df2 = (self.kawa
                        .sheet(sheet_name=unique_id)
                        .select('text', 'record_id', 'int')
                        .compute())
        # then
        self.assertListEqual(list(computed_df1['text']), ['bar', 'foo', 'bar'])
        self.assertListEqual(list(computed_df1['int']), [1, 2, 3])
        self.assertListEqual(list(computed_df1['record_id']), [1, 2, 3])
        self.assertListEqual(list(computed_df2['record_id']), [1, 2, 3, 4, 5, 6, 7, 8, 9])
        self.assertListEqual(list(computed_df2['int']), [1, 2, 3, 1, 2, 3, 1, 2, 3])
        self.assertListEqual(list(computed_df2['text']),
                             ['bar', 'foo', 'bar', 'bar', 'foo', 'bar', 'bar', 'foo', 'bar'])

        # and: check the schemas
        self.assertIn('PRIMARY KEY record_id', schema_1['schema'])
        self.assertIn('ORDER BY record_id', schema_1['schema'])
        self.assertIn('PRIMARY KEY (text, record_id)', schema_2['schema'])
        self.assertIn('ORDER BY (text, record_id)', schema_2['schema'])

    def test_that_new_indicators_are_taken_into_account(self):
        # given
        unique_id = 'resource_{}'.format(uuid.uuid4())

        # when
        loader1 = self.kawa.new_data_loader(
            df=pd.DataFrame([{'dim': 'v1'}, {'dim': 'v2'}]),
            datasource_name=unique_id)
        loader1.create_datasource()
        ds1 = loader1.load_data(create_sheet=True)
        computed_df1 = self.kawa.sheet(sheet_name=unique_id).select('dim', 'record_id').compute()

        # and
        loader2 = self.kawa.new_data_loader(
            df=pd.DataFrame([{'dim': 'v1', 'num': 1}, {'dim': 'v2', 'num': 2}]),
            datasource_name=unique_id)
        loader2.add_new_indicators_to_datasource()
        ds2 = loader2.load_data()
        computed_df2 = self.kawa.sheet(sheet_name=unique_id).select('dim', 'num', 'record_id').compute()

        # then
        self.assertListEqual(list(computed_df1['dim']), ['v1', 'v2'])
        self.assertListEqual(list(computed_df1['record_id']), [1, 2])

        # and
        self.assertListEqual(list(computed_df2['dim']), ['v1', 'v2', 'v1', 'v2'])
        self.assertListEqual(list(computed_df2['record_id']), [1, 2, 3, 4])
        self.assertTrue(math.isnan(list(computed_df2['num'])[0]))
        self.assertTrue(math.isnan(list(computed_df2['num'])[1]))
        self.assertListEqual(list(computed_df2['num'][2:]), [1.0, 2.0])

        # and
        self.assertEqual(len(ds1.get('indicators')), 2)
        self.assertEqual(len(ds2.get('indicators')), 3)

    def test_that_new_indicators_are_taken_into_account_with_existing_pk(self):
        # given
        unique_id = 'resource_{}'.format(uuid.uuid4())

        # when
        loader1 = self.kawa.new_data_loader(
            df=pd.DataFrame([
                {'k': 1, 'dim': 'v1'},
                {'k': 2, 'dim': 'v2'}]),
            datasource_name=unique_id)
        loader1.create_datasource(primary_keys=['k'])
        ds1 = loader1.load_data(create_sheet=True)
        computed_df1 = self.kawa.sheet(sheet_name=unique_id).select('dim').compute()

        # and
        loader2 = self.kawa.new_data_loader(
            df=pd.DataFrame([
                {'k': 3, 'dim': 'v1', 'num': 1},
                {'k': 4, 'dim': 'v2', 'num': 2}
            ]),
            datasource_name=unique_id)
        loader2.add_new_indicators_to_datasource()
        ds2 = loader2.load_data()
        computed_df2 = self.kawa.sheet(sheet_name=unique_id).select('dim', 'num').compute()

        # then
        self.assertListEqual(list(computed_df1['dim']), ['v1', 'v2'])

        # and
        self.assertListEqual(list(computed_df2['dim']), ['v1', 'v2', 'v1', 'v2'])
        self.assertTrue(math.isnan(list(computed_df2['num'])[0]))
        self.assertTrue(math.isnan(list(computed_df2['num'])[1]))
        self.assertListEqual(list(computed_df2['num'][2:]), [1.0, 2.0])

        # and
        self.assertEqual(len(ds1.get('indicators')), 2)
        self.assertEqual(len(ds2.get('indicators')), 3)

    def test_that_new_indicators_are_taken_into_account_when_previous_one_is_removed(self):
        # given
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([{'dim': 'v1'}, {'dim': 'v2'}])
        dtf2 = pd.DataFrame([{'num': 1}, {'num': 2}])

        # when
        loader1 = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader1.create_datasource()
        ds1 = loader1.load_data(create_sheet=True)
        computed_df1 = self.kawa.sheet(sheet_name=unique_id).select('dim', 'record_id').compute()

        # and
        loader2 = self.kawa.new_data_loader(df=dtf2, datasource_name=unique_id)
        loader2.add_new_indicators_to_datasource()
        ds2 = loader2.load_data()
        computed_df2 = self.kawa.sheet(sheet_name=unique_id).select('dim', 'num', 'record_id').compute()

        # then
        self.assertListEqual(list(computed_df1['dim']), ['v1', 'v2'])
        self.assertListEqual(list(computed_df1['record_id']), [1, 2])

        # and
        self.assertListEqual(list(computed_df2['dim']), ['v1', 'v2', '', ''])
        self.assertListEqual(list(computed_df2['record_id']), [1, 2, 3, 4])
        self.assertTrue(math.isnan(list(computed_df2['num'])[0]))
        self.assertTrue(math.isnan(list(computed_df2['num'])[1]))
        self.assertListEqual(list(computed_df2['num'][2:]), [1.0, 2.0])

        # and
        self.assertEqual(len(ds1.get('indicators')), 2)
        self.assertEqual(len(ds2.get('indicators')), 3)

    def test_that_an_exception_is_raised_when_loading_in_a_non_existent_datasource(self):
        # given
        dtf1 = pd.DataFrame([{'dim': 'v1'}, {'dim': 'v2'}])

        # when
        loader1 = self.kawa.new_data_loader(df=dtf1, datasource_name='non existant')
        with self.assertRaises(Exception) as context:
            loader1.load_data(create_sheet=True)

        # then
        self.assertTrue('No datasource with name' in context.exception.args[0])

    def test_that_unicity_with_record_id_is_preserved_with_duplicate_indexes(self):
        # given
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {'index': 1},
            {'index': 1},
            {'index': 1},
            {'index': 1},
            {'index': 1},
            {'index': 1},
            {'index': 1},
            {'index': 1},
            {'index': 1},
            {'index': 1},
            {'index': 1},
            {'index': 1},
        ], index=([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]))

        # when
        loader1 = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader1.create_datasource()
        loader1.load_data(create_sheet=True, nb_threads=2)
        computed_df1 = self.kawa.sheet(sheet_name=unique_id).select('index', 'record_id').compute()

        # then
        self.assertListEqual(list(computed_df1['index']), [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
        self.assertListEqual(list(computed_df1['record_id']), list(range(1, 13)))
